@@ -19,6 +19,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/queue.h>
 
 #include <err.h>
 #include <errno.h>
@@ -32,10 +33,17 @@
 #include <string.h>
 #include <syslog.h>
 #include <time.h>
+#include <tls.h>
 #include <unistd.h>
 
+#include <openssl/evp.h>
+#include <openssl/hmac.h>
+
 #include "got_opentemp.h"
 #include "got_version.h"
+#include "got_object.h"
+
+#include "got_lib_hash.h"
 
 #include "bufio.h"
 #include "log.h"
@@ -91,7 +99,7 @@ dial(const char *host, const char *port)
 
 	freeaddrinfo(res0);
 	if (s == -1)
-		fatal("%s", cause);
+		fatal("%s %s:%s", cause, host, port);
 	return s;
 }
 
@@ -349,7 +357,7 @@ jsonify_commit(FILE *fp, const char *repo, const char *user,
 				break;
 			}
 			phase = P_PARENT;
-			/* fallthough */
+			/* fallthrough */
 
 		case P_PARENT:
 			/* optional - more than one */
@@ -618,7 +626,7 @@ jsonify_tag(FILE *fp, const char *repo, const char *user,
 				break;
 			}
 			phase = P_OBJECT;
-			/* fallthough */
+			/* fallthrough */
 
 		case P_OBJECT:
 			/* optional */
@@ -839,6 +847,60 @@ bufio2poll(struct bufio *bio)
 	return ret;
 }
 
+static unsigned char *
+compute_hmac_sha256(FILE *payload, off_t paylen, const char *hmac_secret,
+    size_t secret_len, unsigned char *hmac_sig_buf, unsigned int *hmac_siglen)
+{
+	HMAC_CTX *ctx;
+	char buf[4096];
+	off_t n;
+	ssize_t r;
+
+	*hmac_siglen = 0;
+
+	ctx = HMAC_CTX_new();
+	if (ctx == NULL) {
+		log_warnx("HMAC_CTX_new failed");
+		return NULL;
+	}
+
+	if (!HMAC_Init_ex(ctx, hmac_secret, secret_len, EVP_sha256(), NULL)) {
+		log_warnx("HMAC_Init_ex failed");
+		goto fail;
+	}
+
+	n = paylen;
+	while (n > 0) {
+		r = fread(buf, 1, n > sizeof(buf) ? sizeof(buf) : n, payload);
+		if (r == 0) {
+			if (feof(payload)) {
+				log_warnx("HMAC payload truncated");
+				goto fail;
+			}
+			log_warn("failed to read HMAC payload");
+			goto fail;
+		}
+		if (!HMAC_Update(ctx, buf, r)) {
+			log_warn("HMAC_Update");
+			goto fail;
+		}
+		n -= r;
+	}
+
+	if (!HMAC_Final(ctx, hmac_sig_buf, hmac_siglen)) {
+		log_warnx("HMAC_Final failed");
+		goto fail;
+	}
+
+	*hmac_siglen = HMAC_size(ctx);
+
+	HMAC_CTX_free(ctx);
+	return hmac_sig_buf;
+fail:
+	HMAC_CTX_free(ctx);
+	return NULL;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -849,11 +911,16 @@ main(int argc, char **argv)
 	const char	*username;
 	const char	*password;
 	const char	*timeoutstr;
+	const char	*hmac_secret;
 	const char	*errstr;
 	const char	*repo = NULL;
 	const char	*host = NULL, *port = NULL, *path = NULL;
 	const char	*gotd_auth_user = NULL;
 	char		*auth, *line, *spc;
+	unsigned char	*hmac_sig = NULL;
+	unsigned char	 hmac_sig_buf[EVP_MAX_MD_SIZE];
+	unsigned int	 hmac_siglen;
+	char		 hex[SHA256_DIGEST_STRING_LENGTH];
 	size_t		 len;
 	ssize_t		 r;
 	off_t		 paylen;
@@ -862,7 +929,14 @@ main(int argc, char **argv)
 	int		 ch, flags, ret, nonstd = 0;
 
 #ifndef PROFILE
-	if (pledge("stdio rpath tmppath dns inet", NULL) == -1)
+	const char *root_ca;
+	root_ca = tls_default_ca_cert_file();
+
+	if (unveil(root_ca, "r") == -1)
+		err(1, "unveil %s", root_ca);
+	if (unveil("/tmp/", "rwc") == -1)
+		err(1, "unveil /tmp");
+	if (pledge("stdio rpath wpath cpath dns inet", NULL) == -1)
 		err(1, "pledge");
 #endif
 
@@ -931,10 +1005,23 @@ main(int argc, char **argv)
 		fatal("fseeko");
 
 #ifndef PROFILE
-	/* drop tmppath */
+	/* drop wpath cpath */
 	if (pledge("stdio rpath dns inet", NULL) == -1)
 		err(1, "pledge");
 #endif
+	hmac_secret = getenv("GOT_NOTIFY_HTTP_HMAC_SECRET");
+	if (hmac_secret) {
+		hmac_sig = compute_hmac_sha256(tmpfp, paylen, hmac_secret,
+		    strlen(hmac_secret), hmac_sig_buf, &hmac_siglen);
+		if (hmac_sig == NULL || hmac_siglen != SHA256_DIGEST_LENGTH)
+			fatalx("HMAC computation failed");
+		if (got_sha256_digest_to_str(hmac_sig, hex, sizeof(hex))
+		    == NULL)
+			fatalx("HMAC conversion to hex string failed");
+
+		if (fseeko(tmpfp, 0, SEEK_SET) == -1)
+			fatal("fseeko");
+	}
 
 	memset(&pfd, 0, sizeof(pfd));
 	pfd.fd = dial(host, port);
@@ -972,10 +1059,15 @@ main(int argc, char **argv)
 	    "Content-Type: application/json\r\n"
 	    "Content-Length: %lld\r\n"
 	    "User-Agent: %s\r\n"
-	    "Connection: close\r\n",
+	    "Connection: close\r\n"
+	    "%s%s%s%s",
 	    path, host,
 	    nonstd ? ":" : "", nonstd ? port : "",
-	    (long long)paylen, USERAGENT);
+	    (long long)paylen, USERAGENT,
+	    hmac_sig ? "X-Gotd-Signature: " : "",
+	    hmac_sig ? "sha256=" : "",
+	    hmac_sig ? hex : "",
+	    hmac_sig ? "\r\n" : "");
 	if (ret == -1)
 		fatal("bufio_compose_fmt");
 
@@ -1038,7 +1130,8 @@ main(int argc, char **argv)
 				fatalx("bad HTTP response from server");
 			*spc++ = '\0';
 			if (strcasecmp(line, "HTTP/1.1") != 0)
-				log_warnx("unexpected protocol: %s", line);
+				log_warnx("warning: unexpected protocol: %s",
+				    line);
 			line = spc;
 
 			spc = strchr(line, ' ');
@@ -1049,7 +1142,7 @@ main(int argc, char **argv)
 			response_code = strtonum(line, 100, 599,
 			    &errstr);
 			if (errstr != NULL)
-				log_warnx("response code is %s: %s",
+				log_warnx("warning: response code is %s: %s",
 				    errstr, line);
 
 			buf_drain(&bio.rbuf, len);
