@@ -494,7 +494,7 @@ spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
     enum gotwebd_proc_type proc_type, const char *username,
     const char *www_user, void (*handler)(int, short, void *))
 {
-	const char	*argv[10];
+	const char	*argv[12];
 	int		 argc = 0;
 	int		 p[2];
 	pid_t		 pid;
@@ -558,6 +558,10 @@ spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
 		argv[argc++] = "-G";
 		argv[argc++] = usernames;
 	}
+	if (strcmp(env->httpd_chroot, D_HTTPD_CHROOT) != 0) {
+		argv[argc++] = "-C";
+		argv[argc++] = env->httpd_chroot;
+	}
 	if (strcmp(env->gotwebd_conffile, GOTWEBD_CONF) != 0) {
 		argv[argc++] = "-f";
 		argv[argc++] = env->gotwebd_conffile;
@@ -933,12 +937,18 @@ main(int argc, char **argv)
 		fatal("%s: calloc", __func__);
 	config_init(env);
 
-	while ((ch = getopt(argc, argv, "A:D:dG:f:F:L:nS:vW:")) != -1) {
+	while ((ch = getopt(argc, argv, "A:C:D:dG:f:F:L:nS:vW:")) != -1) {
 		switch (ch) {
 		case 'A':
 			proc_type = GOTWEBD_PROC_AUTH;
 			get_usernames(&gotwebd_username, &www_username, optarg);
 			break;
+		case 'C':
+			if (strlcpy(env->httpd_chroot, optarg,
+			    sizeof(env->httpd_chroot)) >=
+			    sizeof(env->httpd_chroot))
+				fatalx("chroot path too long");
+			break;
 		case 'D':
 			if (cmdline_symset(optarg) < 0)
 				log_warnx("could not parse macro definition %s",
@@ -1062,6 +1072,8 @@ main(int argc, char **argv)
 		if (chdir("/") == -1)
 			fatal("chdir /");
 
+		log_info("Using chroot: %s", env->httpd_chroot);
+
 		if (setgroups(1, &pw->pw_gid) == -1 ||
 		    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) == -1 ||
 		    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid) == -1)
