@@ -14,12 +14,10 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#define GOT_OBJECT_ID_MAXLEN SHA256_DIGEST_LENGTH
-#define GOT_OBJECT_ID_HEX_MAXLEN SHA256_DIGEST_STRING_LENGTH
-
 enum got_hash_algorithm {
 	GOT_HASH_SHA1,
 	GOT_HASH_SHA256,
+	GOT_NUM_HASH_ALGOS,
 };
 
 struct got_object_id {
