@@ -45,6 +45,10 @@ struct got_raw_object {
 	size_t hdrlen;
 	int refcnt;		/* > 0 if open and/or cached */
 
+	int flags;
+#define GOT_RAW_OBJ_FLAG_PACKED		0x01
+#define GOT_RAW_OBJ_FLAG_DELTIFIED	0x02
+
 	got_object_raw_close_cb *close_cb;
 	void *close_arg;
 };
@@ -155,4 +159,4 @@ const struct got_error *got_object_enumerate(int *,
     struct got_packidx *, struct got_repository *);
 
 const struct got_error *got_object_raw_alloc(struct got_raw_object **,
-    uint8_t *, int *, size_t, off_t);
+    uint8_t *, int *, size_t, size_t, off_t);
