@@ -799,6 +799,7 @@ got_pack_start_privsep_child(struct got_pack *pack, struct got_packidx *packidx)
 	}
 done:
 	if (err) {
+		imsgbuf_clear(ibuf);
 		free(ibuf);
 		free(pack->privsep_child);
 		pack->privsep_child = NULL;
@@ -809,19 +810,20 @@ done:
 static const struct got_error *
 pack_stop_privsep_child(struct got_pack *pack)
 {
-	const struct got_error *err = NULL, *close_err = NULL;
+	const struct got_error *err = NULL;
+	const struct got_error *close_err = NULL, *child_err = NULL;
 
 	if (pack->privsep_child == NULL)
 		return NULL;
 
 	err = got_privsep_send_stop(pack->privsep_child->imsg_fd);
-	if (err)
-		return err;
 	if (close(pack->privsep_child->imsg_fd) == -1)
 		close_err = got_error_from_errno("close");
-	err = got_privsep_wait_for_child(pack->privsep_child->pid);
 	if (close_err && err == NULL)
 		err = close_err;
+	child_err = got_privsep_wait_for_child(pack->privsep_child->pid);
+	if (child_err && err == NULL)
+		err = child_err;
 	imsgbuf_clear(pack->privsep_child->ibuf);
 	free(pack->privsep_child->ibuf);
 	free(pack->privsep_child);
@@ -1627,7 +1629,7 @@ got_pack_dump_delta_chain_to_file(size_t *result_size,
 				free(delta_buf);
 			if (err)
 				goto done;
-			if (fulltext == NULL) {
+			if (pack->delta_cache && fulltext == NULL) {
 				err = got_delta_cache_add_fulltext(
 				    pack->delta_cache, delta->data_offset,
 				    accum_buf, accum_size);
@@ -1889,7 +1891,7 @@ got_pack_dump_delta_chain_to_mem(uint8_t **outbuf, size_t *outlen,
 		if (err)
 			goto done;
 
-		if (fulltext == NULL) {
+		if (pack->delta_cache && fulltext == NULL) {
 			err = got_delta_cache_add_fulltext(pack->delta_cache,
 			    delta->data_offset, accum_buf, accum_size);
 			if (err) {
