@@ -595,6 +595,25 @@ pick_deltas(struct got_pack_meta **meta, int nmeta, int ncolored,
 			n->dtab = NULL;
 		}
 
+		/*
+		 * Do not deltify objects which are already packed and not
+		 * deltified. The object is likely a delta-base for deltas
+		 * we are reusing, and deltifying bases is a waste of time.
+		 * We can just copy them to our new pack file directly.
+		 * In the worst case we could end up using a bit more disk
+		 * space. But our newly created pack will be done sooner.
+		 *
+		 * We cannot skip deltified objects here because we cannot
+		 * tell whether their base object will be included in the
+		 * pack file we are generating.
+		 */
+		if ((raw->flags & GOT_RAW_OBJ_FLAG_PACKED) &&
+		    (raw->flags & GOT_RAW_OBJ_FLAG_DELTIFIED) == 0) {
+			got_object_raw_close(raw);
+			raw = NULL;
+			continue;
+		}
+
 		best_size = raw->size;
 		best_ndeltas = 0;
 		for (j = MAX(0, i - max_base_candidates); j < i; j++) {
@@ -775,7 +794,7 @@ got_pack_load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 {
 	const struct got_error *err;
 	char *p = NULL;
-	int i;
+	int i, ret, psize = PATH_MAX;
 
 	(*ntrees)++;
 	err = got_pack_report_progress(progress_cb, progress_arg, rl,
@@ -783,6 +802,11 @@ got_pack_load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 	if (err)
 		return err;
 
+	p = malloc(psize);
+	if (p == NULL)
+		return got_error_from_errno("malloc");
+	p[0] = '\0';
+
 	for (i = 0; i < got_object_tree_get_nentries(tree); i++) {
 		struct got_tree_entry *e = got_object_tree_get_entry(tree, i);
 		struct got_object_id *id = got_tree_entry_get_id(e);
@@ -806,20 +830,47 @@ got_pack_load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 		if (ids == NULL && S_ISDIR(mode))
 			continue;
 
-		if (asprintf(&p, "%s%s%s", dpath,
+		ret = snprintf(p, psize, "%s%s%s", dpath,
 		    got_path_is_root_dir(dpath) ? "" : "/",
-		    got_tree_entry_get_name(e)) == -1) {
-			err = got_error_from_errno("asprintf");
+		    got_tree_entry_get_name(e));
+		if (ret == -1) {
+			err = got_error_from_errno("snprintf");
 			break;
 		}
+		if (ret >= psize) {
+			char *s;
+
+			s = realloc(p, ret + 1);
+			if (s == NULL) {
+				err = got_error_from_errno("realloc");
+				break;
+			}
+			p = s;
+			psize = ret + 1;
+
+			ret = snprintf(p, psize, "%s%s%s", dpath,
+			    got_path_is_root_dir(dpath) ? "" : "/",
+			    got_tree_entry_get_name(e));
+			if (ret == -1) {
+				err = got_error_from_errno("snprintf");
+				break;
+			}
+			if (ret >= psize) { /* should not happen */
+				err = got_error(GOT_ERR_NO_SPACE);
+				break;
+			}
+		}
 
 		if (S_ISDIR(mode)) {
 			struct got_object_qid *qid;
 			err = got_object_qid_alloc(&qid, id);
 			if (err)
 				break;
-			qid->data = p;
-			p = NULL;
+			qid->data = strdup(p);
+			if (qid->data == NULL) {
+				err = got_error_from_errno("strdup");
+				break;
+			}
 			STAILQ_INSERT_TAIL(ids, qid, entry);
 		} else if (S_ISREG(mode) || S_ISLNK(mode)) {
 			err = got_pack_add_object(want_meta,
@@ -829,11 +880,6 @@ got_pack_load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 			    progress_cb, progress_arg, rl);
 			if (err)
 				break;
-			free(p);
-			p = NULL;
-		} else {
-			free(p);
-			p = NULL;
 		}
 	}
 
@@ -1045,87 +1091,6 @@ got_pack_paint_commit(struct got_object_qid *qid, intptr_t color)
 	return NULL;
 }
 
-const struct got_error *
-got_pack_repaint_parent_commits(struct got_object_id *commit_id, int color,
-    struct got_object_idset *set, struct got_object_idset *skip,
-    struct got_repository *repo)
-{
-	const struct got_error *err = NULL;
-	struct got_object_id_queue ids;
-	struct got_object_qid *qid = NULL;
-	struct got_commit_object *commit = NULL;
-	const struct got_object_id_queue *parents;
-	struct got_object_idset *traversed;
-
-	STAILQ_INIT(&ids);
-
-	traversed = got_object_idset_alloc();
-	if (traversed == NULL)
-		return got_error_from_errno("got_object_idset_alloc");
-
-	err = got_object_open_as_commit(&commit, repo, commit_id);
-	if (err)
-		return err;
-
-	while (commit) {
-		parents = got_object_commit_get_parent_ids(commit);
-		if (parents) {
-			struct got_object_qid *pid;
-			STAILQ_FOREACH(pid, parents, entry) {
-				/*
-				 * No need to traverse parents which are
-				 * already in the desired set or are
-				 * marked for skipping already.
-				 */
-				if (got_object_idset_contains(set, &pid->id))
-					continue;
-				if (skip != set &&
-				    got_object_idset_contains(skip, &pid->id))
-					continue;
-
-				if (got_object_idset_contains(traversed,
-				    &pid->id))
-					continue;
-				err = got_pack_queue_commit_id(&ids, &pid->id,
-				    color, repo);
-				if (err)
-					break;
-				err = got_object_idset_add(traversed,
-				    &pid->id, NULL);
-				if (err)
-					goto done;
-			}
-		}
-		got_object_commit_close(commit);
-		commit = NULL;
-
-		qid = STAILQ_FIRST(&ids);
-		if (qid == NULL)
-			break;
-
-		STAILQ_REMOVE_HEAD(&ids, entry);
-		err = got_object_idset_add(set, &qid->id, NULL);
-		if (err)
-			break;
-
-		err = got_object_open_as_commit(&commit, repo, &qid->id);
-		if (err)
-			break;
-
-		got_object_qid_free(qid);
-		qid = NULL;
-	}
-done:
-	if (commit)
-		got_object_commit_close(commit);
-	if (qid)
-		got_object_qid_free(qid);
-	got_object_id_queue_free(&ids);
-	got_object_idset_free(traversed);
-
-	return err;
-}
-
 const struct got_error *
 got_pack_queue_commit_id(struct got_object_id_queue *ids,
     struct got_object_id *id, intptr_t color, struct got_repository *repo)
@@ -1137,7 +1102,10 @@ got_pack_queue_commit_id(struct got_object_id_queue *ids,
 	if (err)
 		return err;
 
-	STAILQ_INSERT_TAIL(ids, qid, entry);
+	if (color == COLOR_KEEP)
+		STAILQ_INSERT_TAIL(ids, qid, entry);
+	else	
+		STAILQ_INSERT_HEAD(ids, qid, entry);
 	return got_pack_paint_commit(qid, color);
 }
 
@@ -1218,8 +1186,8 @@ got_pack_find_pack_for_commit_painting(struct got_packidx **best_packidx,
 	*best_packidx = NULL;
 
 	/*
-	 * Find the largest pack which contains at least some of the
-	 * commits we are interested in.
+	 * Find the largest pack which contains the commit at the head
+	 * of the queue and some other commits we are interested in.
 	 */
 	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		const char *path_packidx = pe->path;
@@ -1232,15 +1200,24 @@ got_pack_find_pack_for_commit_painting(struct got_packidx **best_packidx,
 			break;
 
 		nobj = be32toh(packidx->hdr.fanout_table[0xff]);
-		if (nobj <= nobj_max)
+		if (nobj < nobj_max)
 			continue;
 
-		STAILQ_FOREACH(qid, ids, entry) {
+		qid = STAILQ_FIRST(ids);
+		idx = got_packidx_get_object_idx(packidx, &qid->id);
+		if (idx == -1)
+			continue;
+		ncommits++;
+
+		qid = STAILQ_NEXT(qid, entry);
+		while (qid) {
 			idx = got_packidx_get_object_idx(packidx, &qid->id);
 			if (idx != -1)
 				ncommits++;
+			qid = STAILQ_NEXT(qid, entry);
 		}
-		if (ncommits > ncommits_max) {
+
+		if (ncommits >= ncommits_max) {
 			best_packidx_path = path_packidx;
 			nobj_max = nobj;
 			ncommits_max = ncommits;
@@ -1289,16 +1266,6 @@ findtwixt(struct got_object_id ***res, int *nres, int *ncolored,
 		goto done;
 	}
 
-	for (i = 0; i < nhead; i++) {
-		struct got_object_id *id = head[i];
-		if (id == NULL)
-			continue;
-		err = queue_commit_or_tag_id(id, COLOR_KEEP, &ids, repo);
-		if (err)
-			goto done;
-		nqueued++;
-	}
-
 	for (i = 0; i < ntail; i++) {
 		struct got_object_id *id = tail[i];
 		if (id == NULL)
@@ -1309,6 +1276,16 @@ findtwixt(struct got_object_id ***res, int *nres, int *ncolored,
 		nqueued++;
 	}
 
+	for (i = 0; i < nhead; i++) {
+		struct got_object_id *id = head[i];
+		if (id == NULL)
+			continue;
+		err = queue_commit_or_tag_id(id, COLOR_KEEP, &ids, repo);
+		if (err)
+			goto done;
+		nqueued++;
+	}
+
 	err = got_pack_paint_commits(ncolored, &ids, nqueued,
 	    keep, drop, skip, repo, progress_cb, progress_arg, rl,
 	    cancel_cb, cancel_arg);
@@ -1355,9 +1332,12 @@ find_pack_for_enumeration(struct got_packidx **best_packidx,
 
 	*best_packidx = NULL;
 
+	if (nids == 0)
+		return NULL;
+
 	/*
-	 * Find the largest pack which contains at least some of the
-	 * commits and tags we are interested in.
+	 * Find the largest pack which contains the commit at the head
+	 * of the queue and some other commits we are interested in.
 	 */
 	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		const char *path_packidx = pe->path;
@@ -1369,15 +1349,20 @@ find_pack_for_enumeration(struct got_packidx **best_packidx,
 			break;
 
 		nobj = be32toh(packidx->hdr.fanout_table[0xff]);
-		if (nobj <= nobj_max)
+		if (nobj < nobj_max)
 			continue;
 
-		for (i = 0; i < nids; i++) {
+		idx = got_packidx_get_object_idx(packidx, ids[0]);
+		if (idx == -1)
+			continue;
+		ncommits++;
+
+		for (i = 1; i < nids; i++) {
 			idx = got_packidx_get_object_idx(packidx, ids[i]);
 			if (idx != -1)
 				ncommits++;
 		}
-		if (ncommits > ncommits_max) {
+		if (ncommits >= ncommits_max) {
 			best_packidx_path = path_packidx;
 			nobj_max = nobj;
 			ncommits_max = ncommits;
