@@ -58,7 +58,7 @@
  * so packed-refs files we write contain a minimal header which disables all
  * packed-refs "traits" supported by Git.
  */
-#define GOT_PACKED_REFS_HEADER	"# pack-refs with:"
+#define GOT_PACKED_REFS_HEADER	"# pack-refs with: "
 
 /* A symbolic reference. */
 struct got_symref {
@@ -204,6 +204,14 @@ parse_ref_file(struct got_reference **ref, const char *name,
 	}
 	if (fstat(fileno(f), &sb) == -1) {
 		err = got_error_from_errno2("fstat", abspath);
+		if (lock)
+			got_lockfile_unlock(lf, -1);
+		goto done;
+	}
+	if (!S_ISREG(sb.st_mode)) {
+		err = got_error(GOT_ERR_NOT_REF);
+		if (lock)
+			got_lockfile_unlock(lf, -1);
 		goto done;
 	}
 
@@ -212,9 +220,7 @@ parse_ref_file(struct got_reference **ref, const char *name,
 		if (feof(f))
 			err = NULL; /* ignore empty files (could be locks) */
 		else {
-			if (errno == EISDIR)
-				err = got_error(GOT_ERR_NOT_REF);
-			else if (ferror(f))
+			if (ferror(f))
 				err = got_ferror(f, GOT_ERR_IO);
 			else
 				err = got_error_from_errno2("getline", abspath);
@@ -643,6 +649,12 @@ got_ref_get_mtime(struct got_reference *ref)
 	return ref->mtime;
 }
 
+time_t
+got_ref_get_cached_committer_time(struct got_reference *ref)
+{
+	return ref->committer_time;
+}
+
 const struct got_error *
 got_ref_cmp_by_name(void *arg, int *cmp, struct got_reference *re1,
     struct got_reference* re2)
@@ -664,8 +676,11 @@ get_committer_time(struct got_reference *ref, struct got_repository *repo)
 	struct got_object_id *id = NULL;
 
 	err = got_ref_resolve(&id, repo, ref);
-	if (err)
+	if (err) {
+		if (err->code == GOT_ERR_NOT_REF && got_ref_is_symbolic(ref))
+			return NULL;
 		return err;
+	}
 
 	err = got_object_get_type(&obj_type, repo, id);
 	if (err)
