@@ -93,8 +93,14 @@ get_reflist_object_ids(struct got_object_id ***ids, int *nobjects,
 		}
 
 		err = got_ref_resolve(&id, repo, re->ref);
-		if (err)
+		if (err) {
+			if (err->code == GOT_ERR_NOT_REF &&
+			    got_ref_is_symbolic(re->ref)) {
+				err = NULL;
+				continue;
+			}
 			goto done;
+		}
 
 		if (wanted_obj_type_mask != GOT_OBJ_TYPE_ANY) {
 			int obj_type;
@@ -139,6 +145,81 @@ done:
 	return err;
 }
 
+static const struct got_error *
+create_temp_packfile(int *packfd, char **tmpfile_path,
+    struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	char *path;
+
+	*packfd = -1;
+
+	if (asprintf(&path, "%s/%s/packing.pack",
+	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR) == -1)
+		return got_error_from_errno("asprintf");
+
+	err = got_opentemp_named_fd(tmpfile_path, packfd, path, "");
+	if (err)
+		goto done;
+
+	if (fchmod(*packfd, GOT_DEFAULT_PACK_MODE) == -1)
+		err = got_error_from_errno2("fchmod", *tmpfile_path);
+done:
+	free(path);
+	if (err) {
+		if (*packfd != -1)
+			close(*packfd);
+		*packfd = -1;
+		free(*tmpfile_path);
+		*tmpfile_path = NULL;
+	}
+	return err;
+}
+
+static const struct got_error *
+install_packfile(FILE **packfile, int *packfd, char **packfile_path,
+    char **tmpfile_path, struct got_object_id *pack_hash,
+    struct got_repository *repo)
+{
+	const struct got_error *err;
+	char *hash_str;
+
+	err = got_object_id_str(&hash_str, pack_hash);
+	if (err)
+		return err;
+
+	if (asprintf(packfile_path, "%s/%s/pack-%s.pack",
+	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR,
+	    hash_str) == -1) {
+		err = got_error_from_errno("asprintf");
+		goto done;
+	}
+
+	if (lseek(*packfd, 0L, SEEK_SET) == -1) {
+		err = got_error_from_errno("lseek");
+		goto done;
+	}
+
+	if (rename(*tmpfile_path, *packfile_path) == -1) {
+		err = got_error_from_errno3("rename", *tmpfile_path,
+		    *packfile_path);
+		goto done;
+	}
+
+	free(*tmpfile_path);
+	*tmpfile_path = NULL;
+
+	*packfile = fdopen(*packfd, "w");
+	if (*packfile == NULL) {
+		err = got_error_from_errno2("fdopen", *packfile_path);
+		goto done;
+	}
+	*packfd = -1;
+done:
+	free(hash_str);
+	return err;
+}
+
 const struct got_error *
 got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
     struct got_reflist_head *include_refs,
@@ -150,8 +231,7 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 	const struct got_error *err = NULL;
 	struct got_object_id **ours = NULL, **theirs = NULL;
 	int nours = 0, ntheirs = 0, packfd = -1, i;
-	char *tmpfile_path = NULL, *path = NULL, *packfile_path = NULL;
-	char *hash_str = NULL;
+	char *tmpfile_path = NULL, *packfile_path = NULL;
 	FILE *delta_cache = NULL;
 	struct got_ratelimit rl;
 
@@ -160,19 +240,9 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 
 	got_ratelimit_init(&rl, 0, 500);
 
-	if (asprintf(&path, "%s/%s/packing.pack",
-	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR) == -1) {
-		err = got_error_from_errno("asprintf");
-		goto done;
-	}
-	err = got_opentemp_named_fd(&tmpfile_path, &packfd, path, "");
+	err = create_temp_packfile(&packfd, &tmpfile_path, repo);
 	if (err)
-		goto done;
-
-	if (fchmod(packfd, GOT_DEFAULT_PACK_MODE) == -1) {
-		err = got_error_from_errno2("fchmod", tmpfile_path);
-		goto done;
-	}
+		return err;
 
 	delta_cache = got_opentemp();
 	if (delta_cache == NULL) {
@@ -212,34 +282,8 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 	if (err)
 		goto done;
 
-	err = got_object_id_str(&hash_str, *pack_hash);
-	if (err)
-		goto done;
-	if (asprintf(&packfile_path, "%s/%s/pack-%s.pack",
-	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR,
-	    hash_str) == -1) {
-		err = got_error_from_errno("asprintf");
-		goto done;
-	}
-
-	if (lseek(packfd, 0L, SEEK_SET) == -1) {
-		err = got_error_from_errno("lseek");
-		goto done;
-	}
-	if (rename(tmpfile_path, packfile_path) == -1) {
-		err = got_error_from_errno3("rename", tmpfile_path,
-		    packfile_path);
-		goto done;
-	}
-	free(tmpfile_path);
-	tmpfile_path = NULL;
-
-	*packfile = fdopen(packfd, "w");
-	if (*packfile == NULL) {
-		err = got_error_from_errno2("fdopen", tmpfile_path);
-		goto done;
-	}
-	packfd = -1;
+	err = install_packfile(packfile, &packfd, &packfile_path,
+	    &tmpfile_path, *pack_hash, repo);
 done:
 	for (i = 0; i < nours; i++)
 		free(ours[i]);
@@ -248,15 +292,14 @@ done:
 		free(theirs[i]);
 	free(theirs);
 	if (packfd != -1 && close(packfd) == -1 && err == NULL)
-		err = got_error_from_errno2("close", packfile_path);
+		err = got_error_from_errno2("close",
+		    packfile_path ? packfile_path : tmpfile_path);
 	if (delta_cache && fclose(delta_cache) == EOF && err == NULL)
 		err = got_error_from_errno("fclose");
 	if (tmpfile_path && unlink(tmpfile_path) == -1 && err == NULL)
 		err = got_error_from_errno2("unlink", tmpfile_path);
 	free(tmpfile_path);
 	free(packfile_path);
-	free(hash_str);
-	free(path);
 	if (err) {
 		free(*pack_hash);
 		*pack_hash = NULL;
@@ -268,8 +311,8 @@ done:
 }
 
 const struct got_error *
-got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
-    struct got_repository *repo,
+got_repo_index_pack(char **idxpath, FILE *packfile,
+    struct got_object_id *pack_hash, struct got_repository *repo,
     got_pack_index_progress_cb progress_cb, void *progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
@@ -279,14 +322,16 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 	int npackfd = -1, idxfd = -1, nidxfd = -1;
 	int tmpfds[3];
 	int idxstatus, done = 0;
+	int nobj_total = 0, nobj_indexed = 0, nobj_loose = 0, nobj_resolved = 0;
 	const struct got_error *err;
 	struct imsgbuf idxibuf;
 	pid_t idxpid;
 	char *tmpidxpath = NULL;
-	char *packfile_path = NULL, *idxpath = NULL, *id_str = NULL;
+	char *packfile_path = NULL, *id_str = NULL;
 	const char *repo_path = got_repo_get_path_git_dir(repo);
 	struct stat sb;
 
+	*idxpath = NULL;
 	memset(&idxibuf, 0, sizeof(idxibuf));
 
 	for (i = 0; i < nitems(tmpfds); i++)
@@ -335,7 +380,7 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		goto done;
 	}
 
-	if (asprintf(&idxpath, "%s/%s/pack-%s.idx",
+	if (asprintf(idxpath, "%s/%s/pack-%s.idx",
 	    repo_path, GOT_OBJECTS_PACK_DIR, id_str) == -1) {
 		err = got_error_from_errno("asprintf");
 		goto done;
@@ -383,8 +428,6 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 	}
 	done = 0;
 	while (!done) {
-		int nobj_total, nobj_indexed, nobj_loose, nobj_resolved;
-
 		if (cancel_cb) {
 			err = cancel_cb(cancel_arg);
 			if (err)
@@ -399,11 +442,18 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		if (nobj_indexed != 0) {
 			err = progress_cb(progress_arg, sb.st_size,
 			    nobj_total, nobj_indexed, nobj_loose,
-			    nobj_resolved);
+			    nobj_resolved, 0);
 			if (err)
 				break;
 		}
 	}
+	if (done) {
+		err = progress_cb(progress_arg, sb.st_size,
+		    nobj_total, nobj_indexed, nobj_loose,
+		    nobj_resolved, done);
+		if (err)
+			goto done;
+	}
 	if (close(imsg_idxfds[0]) == -1) {
 		err = got_error_from_errno("close");
 		goto done;
@@ -413,8 +463,8 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		goto done;
 	}
 
-	if (rename(tmpidxpath, idxpath) == -1) {
-		err = got_error_from_errno3("rename", tmpidxpath, idxpath);
+	if (rename(tmpidxpath, *idxpath) == -1) {
+		err = got_error_from_errno3("rename", tmpidxpath, *idxpath);
 		goto done;
 	}
 	free(tmpidxpath);
@@ -434,7 +484,6 @@ done:
 			err = got_error_from_errno("close");
 	}
 	free(tmpidxpath);
-	free(idxpath);
 	free(packfile_path);
 	return err;
 }
@@ -944,11 +993,19 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			    &qid->id);
 			if (err)
 				goto done;
+			tree_id = got_object_commit_get_tree_id(commit);
 			break;
 		case GOT_OBJ_TYPE_TAG:
 			err = got_object_open_as_tag(&tag, repo, &qid->id);
 			if (err)
 				goto done;
+			/* tree_id will be set below */
+			break;
+		case GOT_OBJ_TYPE_TREE:
+			tree_id = &qid->id;
+			break;
+		case GOT_OBJ_TYPE_BLOB:
+			tree_id = NULL;
 			break;
 		default:
 			/* should not happen */
@@ -956,37 +1013,67 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			goto done;
 		}
 
-		/* Find a tree object to scan. */
-		if (commit) {
-			tree_id = got_object_commit_get_tree_id(commit);
-		} else if (tag) {
+		if (tag) {
+			struct got_object_id *id;
+
 			obj_type = got_object_tag_get_object_type(tag);
+			while (obj_type == GOT_OBJ_TYPE_TAG) {
+				struct got_tag_object *next_tag;
+
+				id = got_object_tag_get_object_id(tag);
+				if (!got_object_idset_contains(traversed_ids,
+				    id)) {
+					err = got_object_idset_add(
+					    traversed_ids, id, NULL);
+					if (err)
+						goto done;
+				}
+
+				err = got_object_open_as_tag(&next_tag, repo,
+				    id);
+				if (err)
+					goto done;
+
+				got_object_tag_close(tag);
+				tag = next_tag;
+				obj_type = got_object_tag_get_object_type(tag);
+			}
+			id = got_object_tag_get_object_id(tag);
 			switch (obj_type) {
 			case GOT_OBJ_TYPE_COMMIT:
+				if (got_object_idset_contains(traversed_ids,
+				    id))
+					break;
+				err = got_object_idset_add(traversed_ids, id,
+				    NULL);
+				if (err)
+					goto done;
 				err = got_object_open_as_commit(&commit, repo,
-				    got_object_tag_get_object_id(tag));
+				    id);
 				if (err)
 					goto done;
 				tree_id = got_object_commit_get_tree_id(commit);
 				break;
 			case GOT_OBJ_TYPE_TREE:
-				tree_id = got_object_tag_get_object_id(tag);
+				tree_id = id;
 				break;
-			default:
-				/*
-				 * Tag points at something other than a
-				 * commit or tree. Leave this weird tag object
-				 * and the object it points to.
-				 */
+			case GOT_OBJ_TYPE_BLOB:
 				if (got_object_idset_contains(traversed_ids,
-				    got_object_tag_get_object_id(tag)))
+				    id))
 					break;
-				err = got_object_idset_add(traversed_ids,
-				    got_object_tag_get_object_id(tag), NULL);
+				err = got_object_idset_add(traversed_ids, id,
+				    NULL);
 				if (err)
 					goto done;
 				break;
+			default:
+				/* should not happen */
+				err = got_error(GOT_ERR_OBJ_TYPE);
+				goto done;
 			}
+		} else if (tree_id == NULL) {
+			/* Blob which has already been marked as traversed. */
+			continue;
 		}
 
 		if (tree_id) {
@@ -994,6 +1081,7 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			    repo, cancel_cb, cancel_arg);
 			if (err)
 				break;
+			tree_id = NULL;
 		}
 
 		if (commit || tag)
@@ -1197,8 +1285,8 @@ done:
 
 static const struct got_error *
 purge_redundant_pack(struct got_repository *repo, const char *packidx_path,
-    int dry_run, int ignore_mtime, time_t max_mtime,
-    int *remove, off_t *size_before, off_t *size_after)
+    int dry_run, int ignore_mtime, time_t max_mtime, int *remove,
+    int has_unref_objects, off_t *size_before, off_t *size_after)
 {
 	static const char *ext[] = {".idx", ".pack", ".rev", ".bitmap",
 	    ".promisor", ".mtimes"};
@@ -1209,19 +1297,21 @@ purge_redundant_pack(struct got_repository *repo, const char *packidx_path,
 	if (strlcpy(path, packidx_path, sizeof(path)) >= sizeof(path))
 		return got_error(GOT_ERR_NO_SPACE);
 
-	/*
-	 * Do not delete pack files which are younger than our maximum
-	 * modification time threshold.  This prevents a race where a
-	 * new pack file which is being added to the repository
-	 * concurrently would be deleted.
-	 */
-	if (fstatat(got_repo_get_fd(repo), path, &sb, 0) == -1) {
-		if (errno == ENOENT)
-			return NULL;
-		return got_error_from_errno2("fstatat", path);
+	if (*remove && !ignore_mtime && has_unref_objects) {
+		/*
+		 * Do not delete pack files which contain unreferenced objects
+		 * and are younger than our modification time threshold.
+		 * This prevents a race where a new pack file which is being
+		 * added to the repository concurrently would be deleted.
+		 */
+		if (fstatat(got_repo_get_fd(repo), path, &sb, 0) == -1) {
+			if (errno == ENOENT)
+				return NULL;
+			return got_error_from_errno2("fstatat", path);
+		}
+		if (sb.st_mtime > max_mtime)
+			*remove = 0;
 	}
-	if (!ignore_mtime && sb.st_mtime > max_mtime)
-		*remove = 0;
 
 	/*
 	 * For compatibility with Git, if a matching .keep file exist
@@ -1269,9 +1359,9 @@ purge_redundant_pack(struct got_repository *repo, const char *packidx_path,
 }
 
 static const struct got_error *
-pack_is_redundant(int *redundant, struct got_repository *repo,
-    struct got_object_idset *traversed_ids,
-    const char *packidx_path, struct got_object_idset *idset)
+pack_is_redundant(int *redundant, int *unref_objects,
+    struct got_repository *repo, struct got_object_idset *traversed_ids,
+    struct got_packidx *large_packidx, const char *packidx_path)
 {
 	const struct got_error *err;
 	struct got_packidx *packidx;
@@ -1279,8 +1369,10 @@ pack_is_redundant(int *redundant, struct got_repository *repo,
 	struct got_object_id id;
 	size_t i, nobjects;
 	size_t digest_len = got_hash_digest_length(repo->algo);
+	int idx;
 
-	*redundant = 1;
+	*redundant = 0;
+	*unref_objects = 0;
 
 	err = got_repo_get_packidx(&packidx, packidx_path, repo);
 	if (err)
@@ -1294,23 +1386,30 @@ pack_is_redundant(int *redundant, struct got_repository *repo,
 		memcpy(&id.hash, pid, digest_len);
 		id.algo = repo->algo;
 
-		if (got_object_idset_contains(idset, &id))
-			continue;
+		/*
+		 * Pack files which contain unknown objects must be kept
+		 * if they meet our modification time threshold.
+		 */
+		if (*unref_objects == 0 &&
+		    !got_object_idset_contains(traversed_ids, &id))
+			*unref_objects = 1;
 
-		if (!got_object_idset_contains(traversed_ids, &id))
-			continue;
-
-		*redundant = 0;
-		err = got_object_idset_add(idset, &id, NULL);
-		if (err)
-			return err;
+		/*
+		 * A pack file is not redundant if any referenced object
+		 * it contains has no valid index in our large pack file,
+		 */
+		idx = got_packidx_get_object_idx(large_packidx, &id);
+		if (idx == -1 && got_object_idset_contains(traversed_ids, &id))
+			return NULL;
 	}
 
+	*redundant = 1;
 	return NULL;
 }
 
 struct pack_info {
 	const char	*path;
+	size_t		 path_len;
 	size_t		 nobjects;
 };
 
@@ -1322,7 +1421,8 @@ pack_info_cmp(const void *a, const void *b)
 	pa = a;
 	pb = b;
 	if (pa->nobjects == pb->nobjects)
-		return strcmp(pa->path, pb->path);
+		return got_path_cmp(pa->path, pb->path,
+		    pa->path_len, pb->path_len);
 	if (pa->nobjects > pb->nobjects)
 		return -1;
 	return 1;
@@ -1332,18 +1432,19 @@ static const struct got_error *
 repo_purge_redundant_packfiles(struct got_repository *repo,
     struct got_object_idset *traversed_ids,
     off_t *size_before, off_t *size_after, int dry_run, int ignore_mtime,
-    time_t max_mtime, int nloose, int ncommits, int npurged,
+    time_t max_mtime, struct got_object_id *large_pack_hash,
+    int nloose, int ncommits, int npurged,
     struct got_ratelimit *rl,
     got_cleanup_progress_cb progress_cb, void *progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err;
 	struct pack_info *pinfo, *sorted = NULL;
-	struct got_packidx *packidx;
-	struct got_object_idset *idset = NULL;
+	struct got_packidx *packidx, *large_packidx = NULL;
 	struct got_pathlist_entry *pe;
 	size_t i, npacks;
-	int remove, redundant_packs = 0;
+	int redundant_packs = 0, large_idxpath_len;
+	char *id_str = NULL, *large_idxpath = NULL;
 
 	npacks = 0;
 	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths)
@@ -1356,6 +1457,16 @@ repo_purge_redundant_packfiles(struct got_repository *repo,
 	if (sorted == NULL)
 		return got_error_from_errno("calloc");
 
+	err = got_object_id_str(&id_str, large_pack_hash);
+	if (err)
+		goto done;
+	large_idxpath_len = asprintf(&large_idxpath, "%s/pack-%s.idx",
+	    GOT_OBJECTS_PACK_DIR, id_str);
+	if (large_idxpath_len == -1) {
+		err = got_error_from_errno("asprintf");
+		goto done;
+	}
+
 	i = 0;
 	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		err = got_repo_get_packidx(&packidx, pe->path, repo);
@@ -1364,32 +1475,45 @@ repo_purge_redundant_packfiles(struct got_repository *repo,
 
 		pinfo = &sorted[i++];
 		pinfo->path = pe->path;
+		pinfo->path_len = pe->path_len;
 		pinfo->nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
 	}
 	qsort(sorted, npacks, sizeof(*sorted), pack_info_cmp);
 
-	idset = got_object_idset_alloc();
-	if (idset == NULL) {
-		err = got_error_from_errno("got_object_idset_alloc");
+	/*
+	 * Open the index of the large pack file we just created.
+	 * Any other pack file composed only of objects which are also
+	 * contained in our large pack is now redundant and can be purged.
+	 */
+	err = got_packidx_open(&large_packidx, got_repo_get_fd(repo),
+	    large_idxpath, 1, repo->algo);
+	if (err)
 		goto done;
-	}
 
 	for (i = 0; i < npacks; ++i) {
+		int is_redundant = 0, has_unref_objects = 0;
+
 		if (cancel_cb) {
 			err = (*cancel_cb)(cancel_arg);
 			if (err)
 				break;
 		}
 
-		err = pack_is_redundant(&remove, repo, traversed_ids,
-		    sorted[i].path, idset);
-		if (err)
-			goto done;
+		/* Do not delete the large pack file we just created. */
+		if (got_path_cmp(sorted[i].path, large_idxpath,
+		    sorted[i].path_len, large_idxpath_len) != 0) {
+			err = pack_is_redundant(&is_redundant,
+			    &has_unref_objects, repo, traversed_ids,
+			    large_packidx, sorted[i].path);
+			if (err)
+				goto done;
+		}
 		err = purge_redundant_pack(repo, sorted[i].path, dry_run,
-		    ignore_mtime, max_mtime, &remove, size_before, size_after);
+		    ignore_mtime, max_mtime, &is_redundant, has_unref_objects,
+		    size_before, size_after);
 		if (err)
 			goto done;
-		if (!remove)
+		if (!is_redundant)
 			continue;
 		err = report_cleanup_progress(progress_cb, progress_arg,
 		    rl, ncommits, nloose, npurged, ++redundant_packs);
@@ -1406,8 +1530,10 @@ repo_purge_redundant_packfiles(struct got_repository *repo,
 	}
  done:
 	free(sorted);
-	if (idset)
-		got_object_idset_free(idset);
+	if (large_packidx)
+		got_packidx_close(large_packidx);
+	free(large_idxpath);
+	free(id_str);
 	return err;
 }
 
@@ -1417,6 +1543,8 @@ got_repo_cleanup(struct got_repository *repo,
     off_t *pack_before, off_t *pack_after,
     int *ncommits, int *nloose, int *npacked, int dry_run, int ignore_mtime,
     got_cleanup_progress_cb progress_cb, void *progress_arg,
+    got_pack_progress_cb pack_progress_cb, void *pack_progress_arg,
+    got_pack_index_progress_cb index_progress_cb, void *index_progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *unlock_err, *err = NULL;
@@ -1427,11 +1555,15 @@ got_repo_cleanup(struct got_repository *repo,
 	struct got_reflist_entry *re;
 	struct got_object_id **referenced_ids;
 	int i, nreferenced;
-	int npurged = 0;
+	int npurged = 0, packfd = -1;
+	char *tmpfile_path = NULL, *packfile_path = NULL, *idxpath = NULL;
+	FILE *delta_cache = NULL, *packfile = NULL;
+	struct got_object_id pack_hash;
 	time_t max_mtime = 0;
 
 	TAILQ_INIT(&refs);
 	got_ratelimit_init(&rl, 0, 500);
+	memset(&pack_hash, 0, sizeof(pack_hash));
 
 	*loose_before = 0;
 	*loose_after = 0;
@@ -1445,6 +1577,16 @@ got_repo_cleanup(struct got_repository *repo,
 	if (err)
 		return err;
 
+	err = create_temp_packfile(&packfd, &tmpfile_path, repo);
+	if (err)
+		goto done;
+
+	delta_cache = got_opentemp();
+	if (delta_cache == NULL) {
+		err = got_error_from_errno("got_opentemp");
+		goto done;
+	}
+
 	traversed_ids = got_object_idset_alloc();
 	if (traversed_ids == NULL) {
 		err = got_error_from_errno("got_object_idset_alloc");
@@ -1469,8 +1611,7 @@ got_repo_cleanup(struct got_repository *repo,
 	}
 
 	err = get_reflist_object_ids(&referenced_ids, &nreferenced,
-	    (1 << GOT_OBJ_TYPE_COMMIT) | (1 << GOT_OBJ_TYPE_TAG),
-	    &refs, repo, cancel_cb, cancel_arg);
+	    GOT_OBJ_TYPE_ANY, &refs, repo, cancel_cb, cancel_arg);
 	if (err)
 		goto done;
 
@@ -1483,6 +1624,28 @@ got_repo_cleanup(struct got_repository *repo,
 			goto done;
 	}
 
+	err = got_pack_create(&pack_hash, packfd, delta_cache,
+	    NULL, 0, referenced_ids, nreferenced, repo, 0,
+	    0, 0, pack_progress_cb, pack_progress_arg,
+	    &rl, cancel_cb, cancel_arg);
+	if (err)
+		goto done;
+
+	err = install_packfile(&packfile, &packfd, &packfile_path,
+	    &tmpfile_path, &pack_hash, repo);
+	if (err)
+		goto done;
+
+	err = got_repo_index_pack(&idxpath, packfile, &pack_hash, repo,
+	    index_progress_cb, index_progress_arg,
+	    cancel_cb, cancel_arg);
+	if (err)
+		goto done;
+
+	err = got_repo_list_packidx(&repo->packidx_paths, repo);
+	if (err)
+		goto done;
+
 	err = repo_purge_unreferenced_loose_objects(repo, traversed_ids,
 	    loose_before, loose_after, *ncommits, nloose, npacked, &npurged,
 	    dry_run, ignore_mtime, max_mtime, &rl, progress_cb, progress_arg,
@@ -1492,11 +1655,17 @@ got_repo_cleanup(struct got_repository *repo,
 
 	err = repo_purge_redundant_packfiles(repo, traversed_ids,
 	    pack_before, pack_after, dry_run, ignore_mtime, max_mtime,
-	    *nloose, *ncommits, npurged, &rl, progress_cb, progress_arg,
-	    cancel_cb, cancel_arg);
+	    &pack_hash, *nloose, *ncommits, npurged,
+	    &rl, progress_cb, progress_arg, cancel_cb, cancel_arg);
 	if (err)
 		goto done;
 
+	if (dry_run) {
+		if (idxpath && unlink(idxpath) == -1)
+			err = got_error_from_errno2("unlink", idxpath);
+		if (packfile_path && unlink(packfile_path) == -1 && err == NULL)
+			err = got_error_from_errno2("unlink", packfile_path);
+	}
  done:
 	if (lk) {
 		unlock_err = got_lockfile_unlock(lk, got_repo_get_fd(repo));
@@ -1505,6 +1674,16 @@ got_repo_cleanup(struct got_repository *repo,
 	}
 	if (traversed_ids)
 		got_object_idset_free(traversed_ids);
+	if (packfd != -1 && close(packfd) == -1 && err == NULL)
+		err = got_error_from_errno2("close",
+		    packfile_path ? packfile_path : tmpfile_path);
+	if (delta_cache && fclose(delta_cache) == EOF && err == NULL)
+		err = got_error_from_errno("fclose");
+	if (tmpfile_path && unlink(tmpfile_path) == -1 && err == NULL)
+		err = got_error_from_errno2("unlink", tmpfile_path);
+	free(tmpfile_path);
+	free(packfile_path);
+	free(idxpath);
 	return err;
 }
 
