@@ -68,7 +68,7 @@ send_raw_obj(struct imsgbuf *ibuf, struct got_object *obj,
 	off_t size;
 	size_t hdrlen;
 
-	if (lseek(fd, SEEK_SET, 0) == -1) {
+	if (lseek(fd, 0L, SEEK_SET) == -1) {
 		err = got_error_from_errno("lseek");
 		goto done;
 	}
@@ -78,7 +78,7 @@ send_raw_obj(struct imsgbuf *ibuf, struct got_object *obj,
 	if (err)
 		goto done;
 
-	err = got_privsep_send_raw_obj(ibuf, size, hdrlen, data);
+	err = got_privsep_send_raw_obj(ibuf, size, hdrlen, 0, data);
 done:
 	free(data);
 	if (close(fd) == -1 && err == NULL)
