@@ -830,6 +830,7 @@ raw_object_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	struct got_object *obj;
 	struct got_object_id id;
 	size_t datalen;
+	int flags = GOT_RAW_OBJ_FLAG_PACKED;
 
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
 	if (datalen != sizeof(iobj))
@@ -855,6 +856,7 @@ raw_object_request(struct imsg *imsg, struct imsgbuf *ibuf,
 		err = got_pack_get_max_delta_object_size(&size, obj, pack);
 		if (err)
 			goto done;
+		flags |= GOT_RAW_OBJ_FLAG_DELTIFIED;
 	} else
 		size = obj->size;
 
@@ -867,7 +869,8 @@ raw_object_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	if (err)
 		goto done;
 
-	err = got_privsep_send_raw_obj(ibuf, obj->size, obj->hdrlen, buf);
+	err = got_privsep_send_raw_obj(ibuf, obj->size, obj->hdrlen,
+	    flags, buf);
 done:
 	free(buf);
 	if (outfile && fclose(outfile) == EOF && err == NULL)
@@ -1284,16 +1287,16 @@ enumerate_tree(int *have_all_entries, struct imsgbuf *ibuf, size_t *totlen,
 			if (!S_ISDIR(pte->mode))
 				continue;
 
+			if (got_object_idset_contains_hash(idset, pte->algo,
+			    pte->id))
+				continue;
+
 			err = got_object_qid_alloc_partial(&eqid);
 			if (err)
 				goto done;
+	
 			eqid->id.algo = pte->algo;
-			memcpy(eqid->id.hash, pte->id, pte->digest_len);
-
-			if (got_object_idset_contains(idset, &eqid->id)) {
-				got_object_qid_free(eqid);
-				continue;
-			}
+			memcpy(&eqid->id.hash, pte->id, pte->digest_len);
 
 			if (asprintf(&p, "%s%s%s", path,
 			    got_path_is_root_dir(path) ? "" : "/",
@@ -1673,140 +1676,13 @@ queue_commit_id(struct got_object_id_queue *ids, struct got_object_id *id,
 		return err;
 
 	memcpy(&qid->id, id, sizeof(qid->id));
-	STAILQ_INSERT_TAIL(ids, qid, entry);
+	if (color == COLOR_KEEP)
+		STAILQ_INSERT_TAIL(ids, qid, entry);
+	else
+		STAILQ_INSERT_HEAD(ids, qid, entry);
 	return paint_commit(qid, color);
 }
 
-static const struct got_error *
-repaint_parent_commits(struct got_object_id *commit_id, int commit_idx,
-    int color, struct got_object_idset *set, struct got_object_idset *skip,
-    struct got_object_id_queue *ids, int *nids,
-    struct got_object_id_queue *painted, int *npainted,
-    struct got_pack *pack, struct got_packidx *packidx,
-    struct got_object_cache *objcache)
-{
-	const struct got_error *err = NULL;
-	const struct got_object_id_queue *parents;
-	struct got_commit_object *commit = NULL;
-	struct got_object_id_queue repaint;
-	struct got_object_idset *traversed;
-
-	STAILQ_INIT(&repaint);
-
-	traversed = got_object_idset_alloc();
-	if (traversed == NULL)
-		return got_error_from_errno("got_object_idset_alloc");
-
-	err = open_commit(&commit, pack, packidx, commit_idx, commit_id,
-	    objcache);
-	if (err)
-		return err;
-
-	while (commit) {
-		struct got_object_qid *pid, *qid;
-		int idx;
-
-		parents = got_object_commit_get_parent_ids(commit);
-		if (parents) {
-			STAILQ_FOREACH(pid, parents, entry) {
-				idx = got_packidx_get_object_idx(packidx,
-				    &pid->id);
-				/*
-				 * No need to traverse parents which are not in
-				 * the pack file, are already in the desired
-				 * set, or are marked for skipping already.
-				 */
-				if (idx == -1)
-					continue;
-				if (got_object_idset_contains(set, &pid->id))
-					continue;
-				if (set != skip &&
-				    got_object_idset_contains(skip, &pid->id))
-					continue;
-
-				if (got_object_idset_contains(traversed,
-				    &pid->id))
-					continue;
-				err = queue_commit_id(&repaint, &pid->id,
-				    color);
-				if (err)
-					goto done;
-				err = got_object_idset_add(traversed,
-				    &pid->id, NULL);
-				if (err)
-					goto done;
-			}
-		}
-		got_object_commit_close(commit);
-		commit = NULL;
-
-		pid = STAILQ_FIRST(&repaint);
-		if (pid == NULL)
-			break;
-
-		err = paint_commit(pid, color);
-		if (err)
-			break;
-
-		err = got_object_idset_add(set, &pid->id, NULL);
-		if (err)
-			break;
-
-		STAILQ_REMOVE_HEAD(&repaint, entry);
-
-		/* Insert or replace this commit on the painted list. */
-		STAILQ_FOREACH(qid, painted, entry) {
-			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
-				continue;
-			err = paint_commit(qid, color);
-			if (err)
-				goto done;
-			got_object_qid_free(pid);
-			pid = qid;
-			break;
-		}
-		if (qid == NULL) {
-			STAILQ_INSERT_TAIL(painted, pid, entry);
-			(*npainted)++;
-		}
-
-		/*
-		 * In case this commit is on the caller's list of
-		 * pending commits to traverse, repaint it there.
-		 */
-		STAILQ_FOREACH(qid, ids, entry) {
-			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
-				continue;
-			err = paint_commit(qid, color);
-			if (err)
-				goto done;
-			break;
-		}
-
-		idx = got_packidx_get_object_idx(packidx, &pid->id);
-		if (idx == -1) {
-			/*
-			 * Should not happen because we only queue
-			 * parents which exist in our pack file.
-			 */
-			err = got_error(GOT_ERR_NO_OBJ);
-			break;
-		}
-
-		err = open_commit(&commit, pack, packidx, idx, &pid->id,
-		    objcache);
-		if (err)
-			break;
-	}
-done:
-	if (commit)
-		got_object_commit_close(commit);
-	got_object_id_queue_free(&repaint);
-	got_object_idset_free(traversed);
-
-	return err;
-}
-
 static const struct got_error *
 paint_commits(struct got_object_id_queue *ids, int *nids,
     struct got_object_idset *keep, struct got_object_idset *drop,
@@ -1873,12 +1749,6 @@ paint_commits(struct got_object_id_queue *ids, int *nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = repaint_parent_commits(&qid->id, idx,
-				    COLOR_SKIP, skip, skip, ids, nids,
-				    &painted, &npainted, pack, packidx,
-				    objcache);
-				if (err)
-					goto done;
 				break;
 			}
 			if (!got_object_idset_contains(keep, &qid->id)) {
@@ -1897,12 +1767,6 @@ paint_commits(struct got_object_id_queue *ids, int *nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = repaint_parent_commits(&qid->id, idx,
-				    COLOR_SKIP, skip, skip, ids, nids,
-				    &painted, &npainted, pack, packidx,
-				    objcache);
-				if (err)
-					goto done;
 				break;
 			}
 			if (!got_object_idset_contains(drop, &qid->id)) {
@@ -2037,9 +1901,8 @@ commit_painting_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	const struct got_error *err = NULL;
 	size_t datalen;
 	struct got_object_id_queue ids;
-	size_t nkeep = 0, ndrop = 0, nskip = 0;
-	int nids = 0;
-	uintptr_t color;
+	size_t nids;
+	int nqueued = 0, done = 0;
 
 	STAILQ_INIT(&ids);
 
@@ -2047,42 +1910,23 @@ commit_painting_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
 
-	color = COLOR_KEEP;
-	err = recv_object_id_queue(&nkeep, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + nkeep > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
+	while (!done) {
+		err = got_privsep_recv_object_id_queue(&done, &ids, &nids, ibuf);
+		if (err)
+			goto done;
+		if (nqueued + nids > INT_MAX) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			goto done;
+		}
+		nqueued += nids;
 	}
-	nids += nkeep;
 
-	color = COLOR_DROP;
-	err = recv_object_id_queue(&ndrop, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + ndrop > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
-	}
-	nids += ndrop;
-
-	color = COLOR_SKIP;
-	err = recv_object_id_queue(&nskip, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + nskip > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
-	}
-	nids += nskip;
-
-	err = paint_commits(&ids, &nids, keep, drop, skip,
+	err = paint_commits(&ids, &nqueued, keep, drop, skip,
 	    pack, packidx, ibuf, objcache);
 	if (err)
 		goto done;
 
-	err = got_privsep_send_painted_commits(ibuf, &ids, &nids, 0, 1);
+	err = got_privsep_send_painted_commits(ibuf, &ids, &nqueued, 0, 1);
 	if (err)
 		goto done;
 
@@ -2142,10 +1986,6 @@ receive_pack(struct got_pack **packp, struct imsgbuf *ibuf)
 		goto done;
 	}
 
-	err = got_delta_cache_alloc(&pack->delta_cache);
-	if (err)
-		goto done;
-
 #ifndef GOT_PACK_NO_MMAP
 	if (pack->filesize > 0 && pack->filesize <= SIZE_MAX) {
 		pack->map = mmap(NULL, pack->filesize, PROT_READ, MAP_PRIVATE,
